use std::{error::Error, fs::File, io::Read, convert::TryInto, collections::HashMap};
use flate2::read::ZlibDecoder;
use crate::reskit::{utility::{Either, get_string, get_u8, skip, get_u32, get_i8, get_i32, get_u16, get_i16}, soundtrack::types::{SampleRate, SampleFormat, PsgEnvelope}};

use super::types::{Sample, Note, PsgSettings};

const DMF_MAGIC_NUMBER: &'static str = ".DelekDefleMask.";
const DMF_SUPPORTED_VERSION: u8 = 0x18;
const DMF_MD: u8 = 0x02;
const DMF_MD_ENHANCED_CH3: u8 = 0x42;

#[derive(Debug)]
pub enum FrameMode {
    Ntsc,
    Pal,
    Custom( u8, u8, u8 )
}

#[derive(Debug, Default)]
pub struct FmOperator {
    am: u8,
    ar: u8,
    dr: u8,
    mult: u8,
    rr: u8,
    sl: u8,
    tl: u8,
    dt2: u8,
    rs: u8,
    dt: i8,
    d2r: u8,
    ssg_mode: u8
}

#[derive(Debug)]
pub struct FmSettings {
    alg: u8,
    fb: u8,
    fms: u8,
    ams: u8,
    operators: [FmOperator; 4]
}

#[derive(Debug)]
pub enum InstrumentType {
    Fm( FmSettings ),
    Psg( PsgSettings )
}

#[derive(Debug)]
pub struct Instrument {
    name: String,
    instrument_type: InstrumentType
}


#[derive(Debug)]
pub struct Effect {
    effect_code: i16,
    effect_value: Option<i16>
}

#[derive(Debug)]
pub struct PatternRow {
    note: Option<Note>,
    volume: Option<i16>,
    effects: Vec<Effect>,
    instrument_index: Option<u16>
}

pub struct DmfModule {
    platform: u8,
    version: u8,
    time_base: u8,
    speed_a: u8,
    speed_b: u8,
    frame_mode: FrameMode,
    rows_per_pattern: u32,
    pattern_matrix: Vec<Vec<u8>>,
    instruments: Vec<Instrument>,
    samples: Vec<Sample>
}

impl DmfModule {

    pub fn from_file( path: &str ) -> Result<DmfModule, Box<dyn Error>> {
        let mut file = File::open( path )?;
        let mut compressed: Vec<u8> = Vec::new();
        file.read_to_end( &mut compressed )?;

        let mut inflator = ZlibDecoder::new( &compressed[..] );
        let mut bytes: Vec<u8> = Vec::new();
        inflator.read_to_end( &mut bytes )?;

        let mut bytes = bytes.iter();

        if get_string( bytes.by_ref(), 16 )? != DMF_MAGIC_NUMBER {
            return Err( format!( "invalid file: missing DMF header" ) )?
        }

        println!( ".DelekDefleMask. DMF file" );

        let version = get_u8( bytes.by_ref() )?;
        if version != DMF_SUPPORTED_VERSION {
            return Err( format!( "invalid file: unsupported version: {}", version ) )?
        }

        let platform = get_u8( bytes.by_ref() )?;
        match platform {
            DMF_MD => println!( "Mode:\tSega Megadrive" ),
            DMF_MD_ENHANCED_CH3 => println!( "Mode:\t\tSega Megadrive (Ext. Ch. 3 Mode)" ),
            _ => return Err( "invalid file: invalid console format" )?
        };

        let song_name_len = get_u8( bytes.by_ref() )?;
        println!( "Song name:\t{}", get_string( bytes.by_ref(), song_name_len.into() )? );

        let song_author_len = get_u8( bytes.by_ref() )?;
        println!( "Song author:\t{}", get_string( bytes.by_ref(), song_author_len.into() )? );

        // Burn 2 bytes, don't care about the highlight patterns
        skip( bytes.by_ref(), 2 );

        // DMF format appears to subtract by 1 from the base time shown in the UI
        // Base time is multiplied by speed_a and speed_b so setting it to 1 and
        // getting 0 doesn't make any sense.
        let time_base = get_u8( bytes.by_ref() )? + 1;
        let speed_a = get_u8( bytes.by_ref() )?;
        let speed_b = get_u8( bytes.by_ref() )?;

        println!( "Time base:\t{}", time_base );
        println!( "Speed A:\t{}", speed_a );
        println!( "Speed B:\t{}", speed_b );

        let frame_mode = match get_u8( bytes.by_ref() )? {
            0 => FrameMode::Pal,
            1 => FrameMode::Ntsc,
            _ => return Err( "invalid file: invalid frame mode" )?
        };

        let custom_frame_mode = get_u8( bytes.by_ref() )? == 1;
        let custom_frame_mode_hz_1 = get_u8( bytes.by_ref() )?;
        let custom_frame_mode_hz_2 = get_u8( bytes.by_ref() )?;
        let custom_frame_mode_hz_3 = get_u8( bytes.by_ref() )?;

        let frame_mode = if custom_frame_mode {
            FrameMode::Custom( custom_frame_mode_hz_1, custom_frame_mode_hz_2, custom_frame_mode_hz_3 )
        } else {
            frame_mode
        };

        println!( "Frame mode:\t{:?}", frame_mode );

        let rows_per_pattern = get_u32( bytes.by_ref() )?;

        println!( "Rows/pattern:\t{}", rows_per_pattern );

        let patterns_count = get_u8( bytes.by_ref() )?;

        println!( "Patterns:\t{}", patterns_count );

        println!( "\nPattern Matrix:" );
        let system_total_channels = if platform == DMF_MD {
            println!( "FM1  FM2  FM3  FM4  FM5  FM6  SN1  SN2  SN3  SN4" );
            10
        } else {
            println!( "FM1  FM2  OP1  OP2  OP3  OP4  FM4  FM5  FM6  SN1  SN2  SN3  SN4" );
            13
        };

        let mut pattern_matrix: Vec<Vec<u8>> = vec![vec![0; system_total_channels]; patterns_count.into()];
        for channel in 0..system_total_channels {
            for pattern in 0..patterns_count {
                let channel: usize = channel.into();
                let pattern: usize = pattern.into();
                pattern_matrix[ pattern ][ channel ] = get_u8( bytes.by_ref() )?;
            }
        }

        for pattern in &pattern_matrix {
            for channel in pattern {
                print!( "{:#04X?} ", channel );
            }
            print!( "\n" );
        }

        print!( "\n" );

        let mut instruments: Vec<Instrument> = Vec::new();
        let instrument_count = get_u8( bytes.by_ref() )?;

        println!( "Instruments:\t{}", instrument_count );

        for i in 0..instrument_count {
            let name_len = get_u8( bytes.by_ref() )?;
            let name = get_string( bytes.by_ref(), name_len.into() )?;

            println!( "Instrument {}:\t{}", i, name );

            let mode = get_u8( bytes.by_ref() )?;
            let instrument_type = match mode {
                0 => {
                    println!( "Psg Instrument" );

                    // Volume envelope
                    let volume = {
                        let size = get_u8( bytes.by_ref() )?;
                        let mut envelope: Vec<u32> = Vec::new();

                        for _i in 0..size {
                            envelope.push( get_u32( bytes.by_ref() )? );
                        }

                        let loop_at: Option<usize> = if size > 0 {
                            let loop_position = get_i8( bytes.by_ref() )?;
                            if loop_position != -1 {
                                Some( loop_position.try_into()? )
                            } else {
                                None
                            }
                        } else {
                            None
                        };

                        println!( "Volume Envelope Size:\t{}", envelope.len() );

                        PsgEnvelope {
                            envelope: Either::Type1( envelope ),
                            loop_at,
                            settings: HashMap::new()
                        }
                    };

                    // Arpeggio envelope
                    let arpeggio = {
                        let size = get_u8( bytes.by_ref() )?;
                        let mut envelope: Vec<i32> = Vec::new();

                        for _i in 0..size {
                            envelope.push( get_i32( bytes.by_ref() )? - 12 );
                        }

                        let loop_at: Option<usize> = if size > 0 {
                            let loop_position = get_i8( bytes.by_ref() )?;
                            if loop_position != -1 {
                                Some( loop_position.try_into()? )
                            } else {
                                None
                            }
                        } else {
                            None
                        };

                        let settings = HashMap::from([
                            ( "arpeggio_macro_fixed", get_u8( bytes.by_ref() )? == 1 )
                        ]);

                        println!( "Arpeggio Envelope Size:\t{}", envelope.len() );
                        println!( "Arpeggio Macro Fixed:\t{}", settings[ "arpeggio_macro_fixed" ] );

                        PsgEnvelope {
                            envelope: Either::Type2( envelope ),
                            loop_at,
                            settings
                        }
                    };

                    // Noise envelope
                    let noise = {
                        let size = get_u8( bytes.by_ref() )?;
                        let mut envelope: Vec<u32> = Vec::new();

                        for _i in 0..size {
                            envelope.push( get_u32( bytes.by_ref() )? );
                        }

                        let loop_at: Option<usize> = if size > 0 {
                            let loop_position = get_i8( bytes.by_ref() )?;
                            if loop_position != -1 {
                                Some( loop_position.try_into()? )
                            } else {
                                None
                            }
                        } else {
                            None
                        };

                        println!( "Noise Envelope Size:\t{}", envelope.len() );

                        PsgEnvelope {
                            envelope: Either::Type1( envelope ),
                            loop_at,
                            settings: HashMap::new()
                        }
                    };

                    // Wavetable envelope
                    let wavetable = {
                        let size = get_u8( bytes.by_ref() )?;
                        let mut envelope: Vec<u32> = Vec::new();

                        for _i in 0..size {
                            envelope.push( get_u32( bytes.by_ref() )? );
                        }

                        let loop_at: Option<usize> = if size > 0 {
                            let loop_position = get_i8( bytes.by_ref() )?;
                            if loop_position != -1 {
                                Some( loop_position.try_into()? )
                            } else {
                                None
                            }
                        } else {
                            None
                        };

                        println!( "WavTbl Envelope Size:\t{}\n", envelope.len() );

                        PsgEnvelope {
                            envelope: Either::Type1( envelope ),
                            loop_at,
                            settings: HashMap::new()
                        }
                    };

                    InstrumentType::Psg( PsgSettings { volume, arpeggio, noise, wavetable } )
                },
                1 => {
                    println!( "Fm Instrument" );

                    let alg = get_u8( bytes.by_ref() )?;
                    let fb = get_u8( bytes.by_ref() )?;
                    let fms = get_u8( bytes.by_ref() )?;
                    let ams = get_u8( bytes.by_ref() )?;

                    println!( "Alg:\t{}", alg );
                    println!( "Fb:\t{}", fb );
                    println!( "FMS:\t{}", fms );
                    println!( "AMS:\t{}", ams );

                    let mut operators: [FmOperator; 4] = [
                        FmOperator::default(),
                        FmOperator::default(),
                        FmOperator::default(),
                        FmOperator::default()
                    ];

                    // Operator 1
                    {
                        operators[ 0 ].am = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].ar = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].dr = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].mult = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].rr = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].sl = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].tl = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].dt2 = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].rs = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].dt = get_i8( bytes.by_ref() )? - 3;
                        operators[ 0 ].d2r = get_u8( bytes.by_ref() )?;
                        operators[ 0 ].ssg_mode = get_u8( bytes.by_ref() )?;

                        println!( "Op 1 AM:\t{}", operators[ 0 ].am );
                        println!( "Op 1 AR:\t{}", operators[ 0 ].ar );
                        println!( "Op 1 DR:\t{}", operators[ 0 ].dr );
                        println!( "Op 1 MULT:\t{}", operators[ 0 ].mult );
                        println!( "Op 1 RR:\t{}", operators[ 0 ].rr );
                        println!( "Op 1 SL:\t{}", operators[ 0 ].sl );
                        println!( "Op 1 TL:\t{}", operators[ 0 ].tl );
                        println!( "Op 1 DT2:\t{}", operators[ 0 ].dt2 );
                        println!( "Op 1 RS:\t{}", operators[ 0 ].rs );
                        println!( "Op 1 DT:\t{}", operators[ 0 ].dt );
                        println!( "Op 1 D2R:\t{}", operators[ 0 ].d2r );
                        println!( "Op 1 SSG_MODE:\t{}\n", operators[ 0 ].ssg_mode );
                    }

                    // Operator 3
                    {
                        operators[ 2 ].am = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].ar = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].dr = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].mult = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].rr = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].sl = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].tl = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].dt2 = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].rs = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].dt = get_i8( bytes.by_ref() )? - 3;
                        operators[ 2 ].d2r = get_u8( bytes.by_ref() )?;
                        operators[ 2 ].ssg_mode = get_u8( bytes.by_ref() )?;

                        println!( "Op 3 AM:\t{}", operators[ 2 ].am );
                        println!( "Op 3 AR:\t{}", operators[ 2 ].ar );
                        println!( "Op 3 DR:\t{}", operators[ 2 ].dr );
                        println!( "Op 3 MULT:\t{}", operators[ 2 ].mult );
                        println!( "Op 3 RR:\t{}", operators[ 2 ].rr );
                        println!( "Op 3 SL:\t{}", operators[ 2 ].sl );
                        println!( "Op 3 TL:\t{}", operators[ 2 ].tl );
                        println!( "Op 3 DT2:\t{}", operators[ 2 ].dt2 );
                        println!( "Op 3 RS:\t{}", operators[ 2 ].rs );
                        println!( "Op 3 DT:\t{}", operators[ 2 ].dt );
                        println!( "Op 3 D2R:\t{}", operators[ 2 ].d2r );
                        println!( "Op 3 SSG_MODE:\t{}\n", operators[ 2 ].ssg_mode );
                    }

                    // Operator 2
                    {
                        operators[ 1 ].am = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].ar = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].dr = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].mult = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].rr = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].sl = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].tl = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].dt2 = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].rs = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].dt = get_i8( bytes.by_ref() )? - 3;
                        operators[ 1 ].d2r = get_u8( bytes.by_ref() )?;
                        operators[ 1 ].ssg_mode = get_u8( bytes.by_ref() )?;

                        println!( "Op 2 AM:\t{}", operators[ 1 ].am );
                        println!( "Op 2 AR:\t{}", operators[ 1 ].ar );
                        println!( "Op 2 DR:\t{}", operators[ 1 ].dr );
                        println!( "Op 2 MULT:\t{}", operators[ 1 ].mult );
                        println!( "Op 2 RR:\t{}", operators[ 1 ].rr );
                        println!( "Op 2 SL:\t{}", operators[ 1 ].sl );
                        println!( "Op 2 TL:\t{}", operators[ 1 ].tl );
                        println!( "Op 2 DT2:\t{}", operators[ 1 ].dt2 );
                        println!( "Op 2 RS:\t{}", operators[ 1 ].rs );
                        println!( "Op 2 DT:\t{}", operators[ 1 ].dt );
                        println!( "Op 2 D2R:\t{}", operators[ 1 ].d2r );
                        println!( "Op 2 SSG_MODE:\t{}\n", operators[ 1 ].ssg_mode );
                    }

                    // Operator 4
                    {
                        operators[ 3 ].am = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].ar = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].dr = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].mult = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].rr = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].sl = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].tl = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].dt2 = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].rs = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].dt = get_i8( bytes.by_ref() )? - 3;
                        operators[ 3 ].d2r = get_u8( bytes.by_ref() )?;
                        operators[ 3 ].ssg_mode = get_u8( bytes.by_ref() )?;

                        println!( "Op 4 AM:\t{}", operators[ 3 ].am );
                        println!( "Op 4 AR:\t{}", operators[ 3 ].ar );
                        println!( "Op 4 DR:\t{}", operators[ 3 ].dr );
                        println!( "Op 4 MULT:\t{}", operators[ 3 ].mult );
                        println!( "Op 4 RR:\t{}", operators[ 3 ].rr );
                        println!( "Op 4 SL:\t{}", operators[ 3 ].sl );
                        println!( "Op 4 TL:\t{}", operators[ 3 ].tl );
                        println!( "Op 4 DT2:\t{}", operators[ 3 ].dt2 );
                        println!( "Op 4 RS:\t{}", operators[ 3 ].rs );
                        println!( "Op 4 DT:\t{}", operators[ 3 ].dt );
                        println!( "Op 4 D2R:\t{}", operators[ 3 ].d2r );
                        println!( "Op 4 SSG_MODE:\t{}\n", operators[ 3 ].ssg_mode );
                    }

                    InstrumentType::Fm( FmSettings { alg, fb, fms, ams, operators } )
                },
                _ => return Err( format!( "invalid file: invalid instrument mode {}", mode ) )?
            };

            instruments.push( Instrument { name, instrument_type } );
        }

        // Wavetables are not used in fm synthesis, this should validate as 0 or error
        let wavetable_count = get_u8( bytes.by_ref() )?;
        if wavetable_count > 0 {
            return Err( "invalid file: wavetables should be zero for system type Sega Megadrive" )?;
        }

        // (channel, patterns)
        let mut channel_patterns: Vec<Vec<PatternRow>> = Vec::new();
        for channel in 0..system_total_channels {
            let mut channel_rows: Vec<PatternRow> = Vec::new();
            let num_effects = get_u8( bytes.by_ref() )?;
            println!( "Patterns for Ch. {}", channel );

            for _ in 0..patterns_count {
                for row_id in 0..rows_per_pattern {
                    let note = get_u16( bytes.by_ref() )?;
                    let octave = get_u16( bytes.by_ref() )?;

                    let note: Option<Note> = if note == 0 && octave == 0 {
                        None
                    } else {
                        Some(
                            match note {
                                100 => Note::NoteOff,
                                1 => Note::CSharp( octave ),
                                2 => Note::D( octave ),
                                3 => Note::DSharp( octave ),
                                4 => Note::E( octave ),
                                5 => Note::F( octave ),
                                6 => Note::FSharp( octave ),
                                7 => Note::G( octave ),
                                8 => Note::GSharp( octave ),
                                9 => Note::A( octave ),
                                10 => Note::ASharp( octave ),
                                11 => Note::B( octave ),
                                12 => Note::C( octave + 1 ), // deflemask y u shift the octave down 1 for C??
                                _ => return Err( format!( "invalid file: invalid note id {}", note ) )?
                            }
                        )
                    };

                    let volume = get_i16( bytes.by_ref() )?;
                    let volume: Option<i16> = if volume == -1 {
                        None
                    } else {
                        Some( volume )
                    };

                    let mut effects: Vec<Effect> = Vec::new();
                    for _ in 0..num_effects {
                        let effect_code = get_i16( bytes.by_ref() )?;
                        let effect_value = get_i16( bytes.by_ref() )?;

                        if effect_code != -1 {
                            effects.push(
                                Effect{
                                    effect_code,
                                    effect_value: if effect_value != -1 { Some( effect_value ) } else { None }
                                }
                            );
                        }
                    }

                    let instrument_index = get_i16( bytes.by_ref() )?;
                    let instrument_index: Option<u16> = if instrument_index != -1 {
                        Some( instrument_index.try_into()? )
                    } else {
                        None
                    };

                    let pattern_row = PatternRow { note, volume, effects, instrument_index };
                    println!( "{}:\t{:?}", row_id, pattern_row );

                    channel_rows.push( pattern_row );
                }
            }

            channel_patterns.push( channel_rows );
        }

        println!( "PCM Samples:" );
        let mut samples: Vec<Sample> = Vec::new();
        let num_samples = get_u8( bytes.by_ref() )?;
        for i in 0..num_samples {
            let sample_size = get_u32( bytes.by_ref() )?;
            let sample_name_chars = get_u8( bytes.by_ref() )?;
            let name = get_string( bytes.by_ref(), sample_name_chars.into() )?;
            let rate = match get_u8( bytes.by_ref() )? {
                1 => SampleRate::Hz8000,
                2 => SampleRate::Hz11025,
                3 => SampleRate::Hz16000,
                4 => SampleRate::Hz22050,
                5 => SampleRate::Hz32000,
                invalid => return Err( format!( "invalid file: invalid sample rate {}", invalid ) )?
            };

            let pitch = get_i8( bytes.by_ref() )? - 5;
            let amp = get_u8( bytes.by_ref() )?;
            let bitrate = match get_u8( bytes.by_ref() )? {
                8 => SampleFormat::Bits8,
                16 => SampleFormat::Bits16,
                invalid => return Err( format!( "invalid file: invalid bitrate {}", invalid ) )?
            };

            let mut sample = Sample { name, rate, pitch, amp, bitrate, data: Vec::new() };

            println!( "{}:\t{:?}", i, sample );

            for _ in 0..sample_size {
                sample.data.push( get_u16( bytes.by_ref() )? );
            }

            samples.push( sample );
        }

        Ok(
            DmfModule {
                platform, version, time_base, speed_a, speed_b, frame_mode, rows_per_pattern,
                pattern_matrix, instruments, samples
            }
         )
    }

    pub fn to_esf( self ) -> Result<Vec<u8>, Box<dyn Error>> {
        // TODO !!
        todo!()
    }

}